[project]
name = "tesseract_runtime"
version = "{{ version }}"
requires-python = ">=3.10"

dependencies = [
{% for dep in runtime_dependencies %}
    "{{ dep }}",
{% endfor %}
]

[project.scripts]
tesseract-runtime = "tesseract_core.runtime.cli:main"

[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"

[tool.hatch.build.targets.wheel]
packages = ["tesseract_core"]
